package djmil.cordacheckers.contracts;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

//import djmil.cordacheckers.states.GameProposalResolutionState;
import djmil.cordacheckers.states.GameProposalState;
import net.corda.v5.base.exceptions.CordaRuntimeException;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.ledger.utxo.Command;
import net.corda.v5.ledger.utxo.StateAndRef;
import net.corda.v5.ledger.utxo.transaction.UtxoLedgerTransaction;

public class GameProposalContract implements net.corda.v5.ledger.utxo.Contract {

    private final static Logger log = LoggerFactory.getLogger(GameProposalContract.class);

    public static enum Action implements Command { 
        CREATE,
        ACCEPT,
        REJECT,
        CANCEL;

        public MemberX500Name getInitiator(GameProposalState gameProposalState) {
            switch (this) {
                case CREATE:
                case CANCEL:
                    return gameProposalState.getIssuer();
                case ACCEPT:
                case REJECT:
                    return gameProposalState.getAcquier();
                default:
                    throw new RuntimeException(UNSUPPORTED_ACTION_VALUE_OF + this.name());
            }
        }

        public MemberX500Name getRespondent(GameProposalState gameProposalState) {
            switch (this) {
                case CREATE:
                case CANCEL:
                    return gameProposalState.getAcquier();
                case ACCEPT:
                case REJECT:
                    return gameProposalState.getIssuer();
                default:
                    throw new RuntimeException(UNSUPPORTED_ACTION_VALUE_OF + this.name());
            }
        }

        public MemberX500Name getSender(StateAndRef<GameProposalState> utxoGameProposal) {
            return getInitiator(utxoGameProposal.getState().getContractState());
        }

        public MemberX500Name getReceiver(StateAndRef<GameProposalState> utxoGameProposal) {
            return getRespondent(utxoGameProposal.getState().getContractState());
        }

        public static final String UNSUPPORTED_ACTION_VALUE_OF = "Unsupported Action value: ";
    }

    @Override
    public void verify(UtxoLedgerTransaction trx) {
        log.info("GameProposalContract.verify() called");

        requireThat(trx.getCommands().size() == 1, REQUIRE_SINGLE_COMMAND);

        switch ((Action)(trx.getCommands().get(0))) {
            case CREATE: {
                requireThat(trx.getInputContractStates().isEmpty(), CREATE_INPUT_STATE);
                requireThat(trx.getOutputContractStates().size() == 1, CREATE_OUTPUT_STATE);

                GameProposalState outputState = trx.getOutputStates(GameProposalState.class).get(0);
                requireThat(outputState != null, CREATE_OUTPUT_STATE);

                requireThat(outputState.getRecipientColor() != null, NON_NULL_RECIPIENT_COLOR);
                break; }

            case ACCEPT:
                throw new CordaRuntimeException("Unimplemented!");
                //break; 

            case REJECT:
                requireThat(trx.getInputContractStates().size() == 1, REJECT_INPUT_STATE);
                requireThat(trx.getOutputContractStates().isEmpty(), REJECT_OUTPUT_STATE);
                requireThat(trx.getInputStates(GameProposalState.class).get(0) != null, REJECT_INPUT_STATE);
                break;

            case CANCEL:
                requireThat(trx.getInputContractStates().size() == 1, CANCEL_INPUT_STATE);
                requireThat(trx.getOutputContractStates().isEmpty(), CANCEL_OUTPUT_STATE);
                requireThat(trx.getInputStates(GameProposalState.class).get(0) != null, CANCEL_INPUT_STATE);
                break;

            default:
                throw new CordaRuntimeException(UNKNOWN_COMMAND);
        }
    }
    
    private void requireThat(boolean asserted, String errorMessage) {
        if(!asserted) {
            throw new CordaRuntimeException("Failed requirement: " + errorMessage);
        }
    }

    static final String REQUIRE_SINGLE_COMMAND = "Require a single command";
    static final String UNKNOWN_COMMAND = "Unsupported command";

    static final String CREATE_INPUT_STATE  = "Create command should have no input states";
    static final String CREATE_OUTPUT_STATE = "Create command should output exactly one GameProposal state";
    static final String NON_NULL_RECIPIENT_COLOR = "GameProposal.recipientColor field can not be null";

    static final String REJECT_INPUT_STATE  = "Reject command should have exactly one GameProposal input state";
    static final String REJECT_OUTPUT_STATE = "Reject command should have no output states";

    static final String CANCEL_INPUT_STATE  = "Cancel command should have exactly one GameProposal input state";
    static final String CANCEL_OUTPUT_STATE = "Cancel command should have no output states";
}
