package djmil.cordacheckers.gameproposal;

import java.util.List;
import java.util.UUID;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import djmil.cordacheckers.FlowResult;
import djmil.cordacheckers.states.GameProposalState;
import net.corda.v5.application.flows.ClientRequestBody;
import net.corda.v5.application.flows.ClientStartableFlow;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.FlowEngine;
import net.corda.v5.application.marshalling.JsonMarshallingService;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.base.exceptions.CordaRuntimeException;
import net.corda.v5.crypto.SecureHash;
import net.corda.v5.ledger.utxo.StateAndRef;
import net.corda.v5.ledger.utxo.UtxoLedgerService;
import net.corda.v5.ledger.utxo.transaction.UtxoSignedTransaction;
import net.corda.v5.ledger.utxo.transaction.UtxoTransactionBuilder;

import java.time.Duration;
import java.time.Instant;

import static djmil.cordacheckers.contracts.GameProposalContract.Action;

public class ActionFlow implements ClientStartableFlow {

    private final static Logger log = LoggerFactory.getLogger(CreateFlow.class);

    @CordaInject
    public JsonMarshallingService jsonMarshallingService;

    @CordaInject
    public UtxoLedgerService ledgerService;

    @CordaInject
    public FlowEngine flowEngine;

    @Override
    @Suspendable
    public String call(ClientRequestBody requestBody) {
        try {
            final ActionFlowArgs args = requestBody.getRequestBodyAs(jsonMarshallingService, ActionFlowArgs.class);
            final Action action = args.getAction();

            final StateAndRef<GameProposalState> inputState = findUnconsumedGameProposalState(args.getGameProposalUuid());

            final UtxoSignedTransaction trx = prepareSignedTransaction(action, inputState);

            final SecureHash trxId = this.flowEngine
                .subFlow( new CommitSubFlow(trx, action.getReceiver(inputState)) );

            return new FlowResult(action+"ED", trxId)  // REJECT+ED
                .toJsonEncodedString(jsonMarshallingService);
        } 
        catch (Exception e) {
            log.warn("GameProposalAction flow failed to process utxo request body " + requestBody + 
                        " because: " + e.getMessage());
            return new FlowResult(e).toJsonEncodedString(jsonMarshallingService);
        }
    }

    @Suspendable
    private StateAndRef<GameProposalState> findUnconsumedGameProposalState (UUID gameProposalUuid) {
        /*
         * Get list of all unconsumed aka 'active' GameProposalStates, then filter by UUID.
         * Note, this is an inefficient way to perform this operation if there are a large
         * number of 'active' GameProposals exists in storage.
         */
        List<StateAndRef<GameProposalState>> stateAndRefs = this.ledgerService
            .findUnconsumedStatesByType(GameProposalState.class)
            .stream()
            .filter(sar -> sar.getState().getContractState().getId().equals(gameProposalUuid))
            .collect(Collectors.toList());

        if (stateAndRefs.size() != 1) {
            throw new CordaRuntimeException("Expected only one GameProposal state with id " + gameProposalUuid +
                                            ", but found " + stateAndRefs.size());
        }
        
        return stateAndRefs.get(0);
    }

    @Suspendable
    private UtxoSignedTransaction prepareSignedTransaction(
        Action action,
        StateAndRef<GameProposalState> inputState
    ) {
        UtxoTransactionBuilder trxBuilder = ledgerService.createTransactionBuilder()
            .setNotary(inputState.getState().getNotaryName())
            .setTimeWindowBetween(Instant.now(), Instant.now().plusMillis(Duration.ofDays(1).toMillis()))
            .addInputState(inputState.getRef())
            .addCommand(action)
            .addSignatories(inputState.getState().getContractState().getParticipants());

        if (action == Action.ACCEPT) {
            // TODO
            // .addOutputState(outputState)
            throw new RuntimeException(action +" unimplemented");
        }

        return trxBuilder.toSignedTransaction();
    }

}
